
#include "llist.h"
#include <string.h>

STATIC INLINE void
new_list(PLIST *ListPtr)
{
	// Allocate list header
	*ListPtr = (PLIST) malloc(sizeof(LIST));
	if (*ListPtr == NULL) {
		LIST_ERROR_PRINT("Error allocating list (%s).\n", strerror(errno));
		exit(-1);
	}	
	memset(*ListPtr, '\0', sizeof(LIST));
	LIST_DEBUG_PRINT("Allocated new list 0x%x.\n", *ListPtr);
}

STATIC INLINE void
new_record(PLIST ListPtr, UINT32 State)
{
	PLIST_ENTRY tempPtr = NULL;
	PCHAR tempPtr2 = NULL;

	// Allocate new record structure	
	LIST_DEBUG_PRINT("Allocating space for new record with state 0x%x.\n", State);
	tempPtr = (PLIST_ENTRY) malloc(sizeof(LIST_ENTRY));
	if (tempPtr == NULL) {
		LIST_ERROR_PRINT("Error allocating list element (%s).\n", strerror(errno));
		exit(-1);
	}
	memset(tempPtr, '\0', sizeof(LIST_ENTRY));
	LIST_DEBUG_PRINT("Allocated space for new record at 0x%x.\n", tempPtr);

	// Special empty list case.
	if (ListPtr->last == NULL) {
		ListPtr->last = tempPtr;
		ListPtr->first = tempPtr;
		tempPtr->next = NULL;
		tempPtr->prev = NULL;
	} else {
		// Insert at end, double linked and looped 
		tempPtr->next = NULL;
		tempPtr->prev = ListPtr->last;
		ListPtr->last->next = tempPtr;	
		ListPtr->last = tempPtr;	
	}

	tempPtr->state = State;
	tempPtr->flags = AllWhiteSpaceFlag;
	(ListPtr->count)++;

	// Allocate initial character string space...hopefully large enough	
	LIST_DEBUG_PRINT("Allocating space for new char string for record 0x%x.\n", ListPtr->last);
	tempPtr2 = (PCHAR) malloc(INITIAL_CHAR_STRING_SIZE + 1);
	if (tempPtr2 == NULL) {
		LIST_ERROR_PRINT("Error allocating char string (%s).\n", strerror(errno));
		exit(-1);
	}
	memset(tempPtr2, '\0', INITIAL_CHAR_STRING_SIZE + 1);
	ListPtr->last->doc = tempPtr2;
	ListPtr->last->max = INITIAL_CHAR_STRING_SIZE;

	LIST_DEBUG_PRINT("Allocated new char string 0x%x number %d size %d.\n", 
	                 ListPtr->last, ListPtr->count, ListPtr->last->max);
}

STATIC INLINE void
new_record_special(PLIST ListPtr, UINT32 State)
{
	PLIST_ENTRY tempPtr = NULL;
	tempPtr = (PLIST_ENTRY) malloc(sizeof(LIST_ENTRY));
	if (tempPtr == NULL) {
		LIST_ERROR_PRINT("Error allocating list element (%s).\n", strerror(errno));
		exit(-1);
	}
	memset(tempPtr, '\0', sizeof(LIST_ENTRY));

	// Special empty list case.
	if (ListPtr->last == NULL) {
		ListPtr->last = tempPtr;
		ListPtr->first = tempPtr;
		tempPtr->next = NULL;
		tempPtr->prev = NULL;
	} else {
		// Insert at end, double linked and looped 
		tempPtr->next = NULL;
		tempPtr->prev = ListPtr->last;
		ListPtr->last->next = tempPtr;	
		ListPtr->last = tempPtr;	
	}
	
	tempPtr->state = State;
	tempPtr->flags = AllWhiteSpaceFlag;
	(ListPtr->count)++;
}

STATIC INLINE int
add_char_entry(PLIST_ENTRY ListEntry, CHAR DocChar)
{
	CHAR tempChar = ListEntry->doc[ListEntry->current-1];

	if (ListEntry == NULL) return(0);

	switch (DocChar) {
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			// Only drop it if it is not in TT mode	
			if (tempChar == ' ' && !(ListEntry->state & TT_Flag)) {
				// Forget it...it's not worth it.... :-)
				return 0;
			} 
			
			// Only convert to space if it is not in TT mode	
			if (ListEntry->state & TT_Flag)
				tempChar = DocChar;
			else tempChar = ' ';
			break;
		default:
			ListEntry->flags &= ~AllWhiteSpaceFlag;
			tempChar = DocChar;
			break;
	}

	// Check to see if this character is going to put us over the limit
	if (ListEntry->current == ListEntry->max) {
		// Need more space!
		PCHAR tempPtr = NULL;
			
		LIST_DEBUG_PRINT("add_char_entry: Need new char buffer --> current [%d] max [%d]\n", 
		                 ListEntry->current, ListEntry->max);

		tempPtr = (PCHAR) malloc(2 * ListEntry->max + 1);
		if (tempPtr == NULL) {
			LIST_ERROR_PRINT("Error reallocating char string (%s).\n", strerror(errno));
			exit(-1);
		}
		memset(tempPtr, '\0', (2 * ListEntry->max + 1));

		// Copy the old info into new buffer and update counters
		memcpy(tempPtr, ListEntry->doc, ListEntry->current);
		ListEntry->max = (2 * ListEntry->max);
	
		// Clean up and store new buffer
		free(ListEntry->doc);
		ListEntry->doc = tempPtr;

		LIST_DEBUG_PRINT("Allocated new char buffer [%s] size [%d]\n", 
		                 ListEntry->doc, ListEntry->max);
	}
	
	// Ok, we have enough space for sure now.
	ListEntry->doc[ListEntry->current] = tempChar;
	(ListEntry->current)++;
	
	//LIST_DEBUG_PRINT("New char buffer [%s]\n", ListEntry->doc);
	return 1;
}

STATIC INLINE int
add_char(PLIST ListPtr, CHAR DocChar)
{
	return (add_char_entry(ListPtr->last, DocChar));
}

STATIC INLINE void
add_char_entry_prepend(PLIST_ENTRY ListEntry, CHAR DocChar)
{
	CHAR tempChar = ListEntry->doc[ListEntry->current-1];

	switch (DocChar) {
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			break;
		default:
			ListEntry->flags &= ~AllWhiteSpaceFlag;
			break;
	}
	tempChar = DocChar;

	// Check to see if this character is going to put us over the limit
	if (ListEntry->current == ListEntry->max) {
		// Need more space!
		PCHAR tempPtr = NULL;
			
		LIST_DEBUG_PRINT("add_char_entry_prepend: Need new char buffer --> current [%d] max [%d]\n", 
		                 ListEntry->current, ListEntry->max);

		tempPtr = (PCHAR) malloc(2 * ListEntry->max + 1);
		if (tempPtr == NULL) {
			LIST_ERROR_PRINT("Error reallocating char string (%s).\n", strerror(errno));
			exit(-1);
		}
		memset(tempPtr, '\0', (2 * ListEntry->max + 1));

		// Copy the old info into new buffer and update counters
		memcpy(&tempPtr[1], ListEntry->doc, ListEntry->current);
		ListEntry->max = (2 * ListEntry->max);
	
		// Clean up and store new buffer
		free(ListEntry->doc);
		ListEntry->doc = tempPtr;

		LIST_DEBUG_PRINT("Allocated new char buffer [%s] size [%d]\n", 
		                 ListEntry->doc, ListEntry->max);
	} else {
		memmove(&ListEntry->doc[1], ListEntry->doc, ListEntry->current);
	}
	
	// Ok, we have enough space for sure now.
	ListEntry->doc[0] = tempChar;
	(ListEntry->current)++;
	
	//LIST_DEBUG_PRINT("New char buffer [%s]\n", ListEntry->doc);
}


STATIC INLINE void
add_string_entry(PLIST_ENTRY ListEntry, CHAR *DocChars, UINT32 Count, UINT32 allws)
{
	// Turn off the all white space flag if we aren't adding all ws.
	if (!allws)
		ListEntry->flags &= ~AllWhiteSpaceFlag;

	// Check to see if these characters are going to put us over the limit
	if ((ListEntry->current + Count) >= ListEntry->max) {
		// Need more space!
		PCHAR tempPtr = NULL;
			
		LIST_DEBUG_PRINT("add_string_entry: Need new char buffer --> current [%d] max [%d]\n", 
		                 ListEntry->current, ListEntry->max);

		tempPtr = (PCHAR) malloc(ListEntry->max + Count * 2 + 1);
		if (tempPtr == NULL) {
			LIST_ERROR_PRINT("Error reallocating char string (%s).\n", strerror(errno));
			exit(-1);
		}
		memset(tempPtr, '\0', (ListEntry->max + Count * 2 + 1));

		// Copy the old info into new buffer and update counters
		memcpy(tempPtr, ListEntry->doc, ListEntry->current);
		ListEntry->max += Count * 2;
	
		// Clean up and store new buffer
		free(ListEntry->doc);
		ListEntry->doc = tempPtr;

		LIST_DEBUG_PRINT("Allocated new char buffer [%s] size [%d]\n", 
		                 ListEntry->doc, ListEntry->max);
	}
	
	// Ok, we have enough space for sure now.
	if (Count != 0) memcpy(&ListEntry->doc[ListEntry->current], DocChars, Count);
	ListEntry->current += Count;
	
	//LIST_DEBUG_PRINT("New char buffer [%s]\n", ListEntry->doc);
}


STATIC INLINE void
add_string(PLIST ListPtr, CHAR *DocChars, UINT32 Count, UINT32 allws)
{
	add_string_entry(ListPtr->last, DocChars, Count, allws);
}

STATIC INLINE void
add_string_entry_prepend(PLIST_ENTRY ListEntry, CHAR *DocChars, UINT32 Count, UINT32 allws)
{
	// Turn off the all white space flag if we aren't adding all ws.
	if (!allws)
		ListEntry->flags &= ~AllWhiteSpaceFlag;

	// Check to see if these characters are going to put us over the limit
	if ((ListEntry->current + Count) >= ListEntry->max) {
		// Need more space!
		PCHAR tempPtr = NULL;
			
		LIST_DEBUG_PRINT("add_string_entry_prepend: Need new char buffer --> current [%d] max [%d]\n", 
		                 ListEntry->current, ListEntry->max);

		tempPtr = (PCHAR) malloc(ListEntry->max + (Count * 2) + 1);
		if (tempPtr == NULL) {
			LIST_ERROR_PRINT("Error reallocating char string (%s).\n", strerror(errno));
			exit(-1);
		}
		memset(tempPtr, '\0', (ListEntry->max + (Count * 2) + 1));

		// Copy the old info into new buffer and update counters
		memcpy(&tempPtr[Count], ListEntry->doc, ListEntry->current);
		ListEntry->max += (2 * Count);
	
		// Clean up and store new buffer
		free(ListEntry->doc);
		ListEntry->doc = tempPtr;

		LIST_DEBUG_PRINT("Allocated new char buffer [%s] size [%d]\n", 
		                 ListEntry->doc, ListEntry->max);
	} else {
		memmove(&ListEntry->doc[Count], ListEntry->doc, ListEntry->current);
	}
	
	// Ok, we have enough space for sure now.
	memcpy(&ListEntry->doc[0], DocChars, Count);
	ListEntry->current += Count;
}


STATIC INLINE void
remove_char(PLIST_ENTRY ListEntry, UINT32 Index)
{
 	int index;

	if (ListEntry == NULL) return;

	for(index=Index; index<ListEntry->current; index++) {
		ListEntry->doc[index] = ListEntry->doc[index+1];
	}
	
	(ListEntry->current)--;
}

STATIC INLINE void
add_state(PLIST ListPtr, UINT32 State)
{
	// OR in the new state to the end of the list
	// Except for color and size.  We need to replace
	if (State & COLOR_FLAG_MASK)
		ListPtr->last->state = ((ListPtr->last->state & ~COLOR_FLAG_MASK)| State);
	else if (State & SIZE_FLAG_MASK)
		ListPtr->last->state = ((ListPtr->last->state & ~SIZE_FLAG_MASK)| State);
	else 
		ListPtr->last->state = (ListPtr->last->state | State);

	LIST_DEBUG_PRINT("Added state 0x%x to get state 0x%x\n", 
	                 State, ListPtr->last->state);
}

STATIC INLINE void
add_flag(PLIST ListPtr, UINT32 Flag)
{
	ListPtr->last->flags = (ListPtr->last->flags | Flag);

//	fprintf(stderr, "Added Flag 0x%x to get flag 0x%x on 0x%x\n", 
//	                 Flag, ListPtr->last->flags, ListPtr->last);
}

STATIC INLINE void
remove_flag(PLIST ListPtr, UINT32 Flag)
{
	ListPtr->last->flags &= ~Flag;

//	fprintf(stderr,"Removed Flag 0x%x to get flag 0x%x on 0x%x\n", 
//	                 Flag, ListPtr->last->flags, ListPtr->last);
}

STATIC INLINE void
remove_state(PLIST_ENTRY ListEntry, UINT32 State)
{
	if (ListEntry != NULL) 
		ListEntry->state = (ListEntry->state & ~State);
}

STATIC INLINE void
release_record(PLIST ListPtr, PLIST_ENTRY ListEntry)
{
	if (ListEntry == NULL) return;
	
	LIST_DEBUG_PRINT("Releasing list 0x%x record 0x%x now only %d records.\n", 
	                 ListPtr, ListEntry, ListPtr->count);

	// Special case of only one in list.
	if ((ListPtr->first == ListEntry) && (ListPtr->last == ListEntry)) {
		ListPtr->first = NULL;
		ListPtr->last = NULL;
		ListPtr->count = 0;
		if (ListEntry->doc != NULL) free(ListEntry->doc);
		free(ListEntry);
		return;
	}

	if (ListPtr->first == ListEntry) {
		// Special case of first in list
		ListPtr->first->next->prev = NULL;
		ListPtr->first = ListEntry->next;
	} else if (ListPtr->last == ListEntry) {
		// Special case of last in list
		ListPtr->last->prev->next = NULL;
		ListPtr->last = ListEntry->prev;
	} else {
		// Normal case
		ListEntry->next->prev = ListEntry->prev;
		ListEntry->prev->next = ListEntry->next;
	}

	(ListPtr->count)--;
	LIST_DEBUG_PRINT("Released record 0x%x now only %d records.\n", 
	                 ListEntry, ListPtr->count);

	if (ListEntry->doc != NULL) free(ListEntry->doc);
	free(ListEntry);
}

STATIC INLINE void
release_list(PLIST ListPtr)
{
	while(ListPtr->first != NULL) {
		release_record(ListPtr, ListPtr->last);
	}
	
	free(ListPtr);
}

STATIC INLINE UINT32
get_previous_state(PLIST_ENTRY ListEntry)
{
	if (ListEntry->prev != NULL)
		return(ListEntry->prev->state);
	else
		return(0x0);
}

STATIC INLINE UINT32
get_this_state(PLIST_ENTRY ListEntry)
{
	if (ListEntry != NULL)
		return(ListEntry->state);
	else
		return(0x0);
}

STATIC INLINE UINT32
get_next_state(PLIST_ENTRY ListEntry)
{
	if (ListEntry->next != NULL)
		return(ListEntry->next->state);
	else
		return(0x0);
}

STATIC INLINE UINT32
get_previous_several_states(PLIST_ENTRY ListEntry, UINT32 Count, UINT32 *Storage)
{
	PLIST_ENTRY tempPtr = ListEntry->prev;
	UINT32 counter = 0;

	while (tempPtr != NULL) {
		Storage[0] = tempPtr->state;
		counter++;
		tempPtr = tempPtr->prev;
	}

	return(counter);
}

STATIC INLINE UINT32
get_next_several_states(PLIST_ENTRY ListEntry, UINT32 Count, UINT32 *Storage)
{
	PLIST_ENTRY tempPtr = ListEntry->next;
	UINT32 counter = 0;

	while (tempPtr != NULL) {
		Storage[0] = tempPtr->state;
		counter++;
		tempPtr = tempPtr->next;
	}

	return(counter);
}

STATIC INLINE PLIST_ENTRY
get_first_entry(PLIST ListPtr)
{
	if (ListPtr != NULL) 
		return(ListPtr->first);
	else
		return((PLIST_ENTRY) NULL);
}

STATIC INLINE PLIST_ENTRY
get_last_entry(PLIST ListPtr)
{
	if (ListPtr != NULL) 
		return(ListPtr->last);
	else
		return((PLIST_ENTRY) NULL);
}

STATIC INLINE PLIST_ENTRY
get_next_entry(PLIST_ENTRY ListEntry)
{
	if (ListEntry != NULL) 
		return(ListEntry->next);
	else
		return((PLIST_ENTRY) NULL);
}

STATIC INLINE PLIST_ENTRY
get_previous_entry(PLIST_ENTRY ListEntry)
{
	if (ListEntry != NULL) 
		return(ListEntry->prev);
	else
		return((PLIST_ENTRY) NULL);
}

STATIC INLINE PCHAR
get_previous_string(PLIST_ENTRY ListEntry)
{
	if (ListEntry->prev != NULL) 
		return(ListEntry->prev->doc);
	else
		return((PCHAR) NULL);
}

STATIC INLINE PCHAR
get_this_string(PLIST_ENTRY ListEntry)
{
	if (ListEntry != NULL) 
		return(ListEntry->doc);
	else
		return((PCHAR) NULL);
}
	
STATIC INLINE PCHAR
get_next_string(PLIST_ENTRY ListEntry)
{
	if (ListEntry->next != NULL) 
		return(ListEntry->next->doc);
	else
		return((PCHAR) NULL);
}

STATIC INLINE UINT32
get_previous_size(PLIST_ENTRY ListEntry) 
{
	if (ListEntry->prev != NULL) 
		return(ListEntry->prev->current);
	else
		return(0x0);
}

STATIC INLINE UINT32
get_this_size(PLIST_ENTRY ListEntry) 
{
	if (ListEntry != NULL) 
		return(ListEntry->current);
	else
		return(0x0);
}

STATIC INLINE UINT32
get_next_size(PLIST_ENTRY ListEntry) 
{
	if (ListEntry->next != NULL) 
		return(ListEntry->next->current);
	else
		return(0x0);
}


STATIC INLINE void
print_entry(PLIST_ENTRY ListEntry)
{
	fprintf(stderr, "ENTRY 0x%08x:\n", ListEntry);
	fprintf(stderr, "\tstate   = 0x%08x\n", ListEntry->state);
	fprintf(stderr, "\tflags   = 0x%08x\n", ListEntry->flags);
	fprintf(stderr, "\tcurrent = %d\n", ListEntry->current);
	fprintf(stderr, "\tmax     = %d\n", ListEntry->max);
	fprintf(stderr, "\tnext    = 0x%08x\n", ListEntry->next);
	fprintf(stderr, "\tprev    = 0x%08x\n", ListEntry->prev);
	fprintf(stderr, "\tdoc     = [%s]\n", ListEntry->doc);
}

STATIC INLINE void
print_list(PLIST ListPtr)
{
	PLIST_ENTRY tempPtr = ListPtr->first;
	
	if (ListPtr == NULL) return;
	
	fprintf(stderr, "HEADER:\n");
	fprintf(stderr, "\tcount = %d\n", ListPtr->count);
	fprintf(stderr, "\tfirst = 0x%08x\n", ListPtr->first);
	fprintf(stderr, "\tlast  = 0x%08x\n", ListPtr->last);
	fprintf(stderr, "\n");
	
	while (tempPtr != NULL) {
		print_entry(tempPtr);
		tempPtr = tempPtr->next;
	}
}

STATIC INLINE UINT32
get_max_size(PLIST ListPtr)
{
	PLIST_ENTRY tempPtr = ListPtr->first;
	UINT32 max_size = 0;

	tempPtr = ListPtr->first;
	if (tempPtr == NULL) return(0);

	if (tempPtr->max > max_size) max_size = tempPtr->max;
	tempPtr = tempPtr->next;
	while (tempPtr != NULL) {
		if (tempPtr->max > max_size) max_size = tempPtr->max;
		tempPtr = tempPtr->next;
	}
	
	return(max_size);
}

#if TEST_LIST_AS_APP
int
main(int argc, char *argv[])
{
	PLIST ListPtr = NULL;
	PLIST_ENTRY ListEntry = NULL;
	int index1, index2;
	int numToInsert = 10;
	int numOfChars = 1;
	
	if (argc >= 2) {
		numToInsert = atoi(argv[1]);
	}

	printf("Allocating new list\n");
	new_list(&ListPtr);
	printf("Allocated new list\n");

	for (index1=0; index1<numToInsert; index1++) {
		printf("Inserting record %d into list 0x%x with %d chars.\n", index1, ListPtr, numOfChars);
		
		new_record(ListPtr, 0x00000000);	
		printf("Adding %d chars to record 0x%x...", numOfChars, ListPtr->last);
		for (index2=0; index2<numOfChars; index2++) {
			//printf(".");
			add_char(ListPtr, 'x');
		}
		printf("Done\n");
		add_state(ListPtr, 0x00000001);

		numOfChars += 10;
	}

	print_list(ListPtr);

	release_list(ListPtr);
}
#endif
