#ifndef __LLIST_H__
#define __LLIST_H__

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/errno.h>

// Project types and defs
#include "defs.h"

// Testing
//#define TEST_LIST_AS_APP 1
//#define LIST_DEBUG 0

// Local defines
#if LIST_DEBUG
#	define LIST_DEBUG_PRINT(x, y...) { fprintf(stderr, x, ##y); fflush(stderr); }
#else
#	define LIST_DEBUG_PRINT(x, y...) {}
#endif
#define LIST_ERROR_PRINT(x, y...) { fprintf(stderr, "LIST_ERROR: " x, ##y); fflush(stderr); }

#define INITIAL_CHAR_STRING_SIZE 10

#ifndef DO_INLINING
// List maintanence
void new_list(PLIST *ListPtr);
void new_record(PLIST ListPtr, UINT32 State);
int  add_char(PLIST ListPtr, CHAR DocChar);
int  add_char_entry(PLIST_ENTRY ListEntry, CHAR DocChar);
void add_string(PLIST ListPtr, CHAR *DocChars, UINT32 Count, UINT32 allws);
void add_string_entry(PLIST_ENTRY ListEntry, CHAR *DocChars, UINT32 Count, UINT32 allws);
void add_string_entry_prepend(PLIST_ENTRY ListEntry, CHAR *DocChars, UINT32 Count, UINT32 allws);
void add_state(PLIST ListPtr, UINT32 State);
void add_flag(PLIST ListPtr, UINT32 Flag);
void remove_flag(PLIST ListPtr, UINT32 Flag);
void release_record(PLIST ListPtr, PLIST_ENTRY ListEntry);
void release_list(PLIST ListPtr);
void print_list(PLIST ListPtr);
void print_entry(PLIST_ENTRY ListEntry);
void remove_char(PLIST_ENTRY ListEntry, UINT32 Index);
void remove_state(PLIST_ENTRY ListEntry, UINT32 State);

// State functions
UINT32 get_previous_several_states(PLIST_ENTRY ListEntry, UINT32 Count, UINT32 *Storage);
UINT32 get_previous_state(PLIST_ENTRY ListEntry);
UINT32 get_this_state(PLIST_ENTRY ListEntry);
UINT32 get_next_state(PLIST_ENTRY ListEntry);
UINT32 get_next_several_states(PLIST_ENTRY ListEntry, UINT32 Count, UINT32 *Storage);

// List walking functions
PLIST_ENTRY get_first_entry(PLIST ListPtr);
PLIST_ENTRY get_last_entry(PLIST ListPtr);
PLIST_ENTRY get_previous_entry(PLIST_ENTRY ListEntry);
PLIST_ENTRY get_next_entry(PLIST_ENTRY ListEntry);

// String functions
PCHAR get_previous_string(PLIST_ENTRY ListEntry);
PCHAR get_this_string(PLIST_ENTRY ListEntry);
PCHAR get_next_string(PLIST_ENTRY ListEntry);

// String size functions
UINT32 get_previous_size(PLIST_ENTRY ListEntry);
UINT32 get_this_size(PLIST_ENTRY ListEntry);
UINT32 get_next_size(PLIST_ENTRY ListEntry);

#define STATIC
#define INLINE

#else

#define STATIC	static
#define INLINE	inline

#include "llist.c"

#endif // !DO_INLINING

#endif  // __LLIST_H__
