
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include "llist.h"

extern PLIST head_of_list;

extern int yydebug;
extern void yyparse_init();

FILE *fporig, *fpintermediate, *fpfinal;
UINT32 fporigcount = 0, fpintermediatecount = 0, fpfinalcount = 0, fpmincount = 0;

int debug = 0;

#define MAX_SAFETY_TIME		10

int complete_parsing = 0;
int complete_printing = 0;

void
alarm_handler(int val)
{
	if (complete_printing)
		return;

	if (!complete_parsing || fporigcount <= fpintermediatecount)
	{
		rewind(fporig);
		while (!feof(fporig))
		{
			fputc(fgetc(fporig), stdout);
		}
		exit(0);
	}

	rewind(fpintermediate);
	while (!feof(fpintermediate))
	{
		fputc(fgetc(fpintermediate), stdout);
	}
	exit(0);
}


main( int argc, char **argv )
{
	PLIST_ENTRY entry;
	time_t		the_time;
	int			amount_of_time;
	int         c;
	int         fpflag;

	if (argv[1][0] == '-' && argv[1][1] == 'd' && argv[1][2] == 0)
	{
		debug=1;
		argc--;
		argv++;
	}

	amount_of_time = atoi(argv[1]);
	if (debug)
	{
		fprintf(stderr, "Amount of Time = %d\n", amount_of_time);
	}

	signal(SIGALRM, alarm_handler);
	alarm(MAX_SAFETY_TIME > amount_of_time ? amount_of_time : amount_of_time - MAX_SAFETY_TIME);

	fporig = fopen("/tmp/opter.orig", "w+");
	fpintermediate = fopen("/tmp/opter.intermediate", "w+");
	fpfinal = fopen("/tmp/opter.final", "w+");

	yyparse_init();

	if (debug)
	{
		the_time = time(NULL);
		fprintf(stderr, "Starting Parse %s\n", ctime(&the_time));
	}

	yyparse();
	complete_parsing = 1;

	yyparse_complete();

	if (debug)
	{
		the_time = time(NULL);
		fprintf(stderr, "Done Parsing %s\n", ctime(&the_time));

		// print list
		if (debug)
			print_list(head_of_list);
	}

	// Walk list and merge spaces and neighbors
	entry = get_first_entry(head_of_list);
	while (entry && entry != head_of_list->last)
	{
		PLIST_ENTRY next_entry = get_next_entry(entry);
		UINT32 bstate = get_this_state(entry);
		UINT32 nbstate = get_this_state(next_entry);
		UINT32 state = bstate & WHITE_SPACE_EFFECT_MASK;
		UINT32 nstate = nbstate & WHITE_SPACE_EFFECT_MASK;
		UINT32 ustate = state & UNDERLINE_FLAG_MASK;
		UINT32 unstate = nstate & UNDERLINE_FLAG_MASK;

		if (entry->current == 0) {
			release_record(head_of_list, entry);
			entry = next_entry;
			continue;
		}

		// Try to merge a single whitespace
		if (entry->flags & AllWhiteSpaceFlag)
		{
			if (next_entry->current >= 1 &&
				((state == nstate) || 
				 (((state & (TT_Flag|SIZE_FLAG_MASK)) == (nstate & (TT_Flag|SIZE_FLAG_MASK))) &&
				  (ustate == 0) && (unstate == 0))))
			{
				if (bstate & TT_Flag || next_entry->doc[0] != ' ')
					add_string_entry_prepend(next_entry, entry->doc, entry->current, entry->flags & AllWhiteSpaceFlag);
					next_entry->flags |= (entry->flags & ~AllWhiteSpaceFlag);
				if (head_of_list->first != entry)
					next_entry = entry->prev;
				release_record(head_of_list, entry);
				entry = next_entry;
				continue;
			}
		}

		// Try to merge a single whitespace
		if (next_entry->flags & AllWhiteSpaceFlag)
		{
			if (entry->current >= 1 &&
				((state == nstate) || 
				 (((state & (TT_Flag|SIZE_FLAG_MASK)) == (nstate & (TT_Flag|SIZE_FLAG_MASK))) &&
				  (ustate == 0) && (unstate == 0))))
			{
				if (bstate & TT_Flag || entry->doc[entry->current-1] != ' ')
					add_string_entry(entry, next_entry->doc, next_entry->current, next_entry->flags & AllWhiteSpaceFlag);
				if (entry->next && entry->next->next)
					entry->next->next->flags |= (next_entry->flags & ~AllWhiteSpaceFlag);
				release_record(head_of_list, next_entry);
				continue;
			}
		}

		if (get_this_state(entry) == get_this_state(next_entry))
		{
			add_string_entry(entry, next_entry->doc, next_entry->current, next_entry->flags & AllWhiteSpaceFlag);
			if (entry->next && entry->next->next)
				entry->next->next->flags |= (next_entry->flags & AllWhiteSpaceFlag);
			release_record(head_of_list, next_entry);
			continue;
		}

		// Try to trim a whitespace off one of two consecutive parts
		if (entry->current > 1 && next_entry->current > 1)
		{
			if (entry->current >= 1 && entry->doc[entry->current-1] == ' ' &&
				next_entry->current >= 1 && next_entry->doc[0] == ' ' &&
				((state == nstate) || 
				 (((state & (TT_Flag|SIZE_FLAG_MASK)) == (nstate & (TT_Flag|SIZE_FLAG_MASK))) &&
				  (ustate == 0) && (unstate == 0))))
			{
				if (!(nbstate & TT_Flag))
					remove_char(next_entry, 0);
			}
		}

		entry = next_entry;
	}

	if (head_of_list->last && head_of_list->last->current == 0)
		release_record(head_of_list, get_last_entry(head_of_list));

	if (debug)
	{
		the_time = time(NULL);
		fprintf(stderr, "Done white spacing %s\n", ctime(&the_time));

		// print list
		if (debug)
			print_list(head_of_list);
	}

	print_function(head_of_list);
	complete_printing = 1;

	alarm(0);

	if (debug)
	{
		the_time = time(NULL);
		fprintf(stderr, "Done Setting up printer%s\n", ctime(&the_time));
	}

	// GREG: Really print

	if (debug)
	{
		the_time = time(NULL);
		fprintf(stderr, "Done %s\n", ctime(&the_time));
	}

	fclose(fporig);
	fclose(fpintermediate);
	fclose(fpfinal);

	fpmincount = fporigcount;	
	fpflag = 1;

	if (fpintermediatecount < fpmincount) {
		fpmincount = fpintermediatecount;
		fpflag = 2;
	}

	if (fpfinalcount < fpmincount) {
		fpmincount = fpintermediatecount;
		fpflag = 3;
	}
	
	switch(fpflag) {
		case 1:
			fpfinal = fopen("/tmp/opter.orig", "r");
			break;
		case 2:
			fpfinal = fopen("/tmp/opter.intermediate", "r");
			break;
		case 3:
			fpfinal = fopen("/tmp/opter.final", "r");
			break;
		default:
			fpfinal = fopen("/tmp/opter.final", "r");
			break;
	}
	
	while ((c = fgetc(fpfinal)) != EOF) {
		fputc(c, stdout);
	}
	fclose(fpfinal);

	//fprintf(stderr, "Choice %d: %d %d %d\n", fpflag, fporigcount, fpintermediatecount, fpfinalcount);

	exit(0);
}
